eval 'exec perl -x $0 ${1+"$@"}' # -*-perl-*-
  if 0;
#!perl -w
#
# ======================================================================
# This file is Copyright 1998,1999 by the Purdue Research Foundation and
# may only be used under license.  For terms of the license, see the
# file named COPYRIGHT included with this software release.
# AAFID is a trademark of the Purdue Research Foundation.
# All rights reserved.
# ======================================================================
#
# AAFID project, COAST Laboratory, CERIAS, 1998-1999.
#
# Package AAFID::System
#
# Diego Zamboni, Jan 27, 1998.
#
# $Id: System.pm,v 1.2 1999/09/03 17:08:54 zamboni Exp $
#
# NOTE: This file is in Perl's POD format. For more information, see the 
#       manual page for perlpod(1).
#

package AAFID::System;

# The following keeps up with the RCS version number. The self-assignment
# keeps the -w switch from complaining by using the variable.
$VERSION = do { my @r = (q$Revision: 1.2 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r }; $VERSION = $VERSION;

=pod

This module contains routines that act differently depending on the
operating system (and probably architecture) on which we are running.

=head1 Temporary directory

Return a suitable default temporary directory.

=cut

sub tmpdir {
  my $osname=$^O;
  # TODO: Check if this is the appropriate string for windows.
  if ($osname eq "windows") {
    return "c:\tmp";
  }
  else {
    # Assume anything else is Unix for now
    return "/tmp";
  }
}

1;

