# The following keeps up with the RCS version number. The self-assignment
# keeps the -w switch from complaining by using the variable.
$VERSION = do { my @r = (q$Revision: 1.3 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r }; $VERSION = $VERSION;
