eval 'exec perl -x $0 ${1+"$@"}' # -*-perl-*-
  if 0;
#!perl -w
#
# ======================================================================
# This file is Copyright 1998,1999 by the Purdue Research Foundation and
# may only be used under license.  For terms of the license, see the
# file named COPYRIGHT included with this software release.
# AAFID is a trademark of the Purdue Research Foundation.
# All rights reserved.
# ======================================================================
#
# AAFID project, COAST Laboratory, CERIAS, 1998-1999.
#
# Frederic Dumont, 1998.
#
# $Id: Tags.pm,v 1.4 1999/09/03 16:49:51 zamboni Exp $
#
# NOTE: This file is in Perl's POD format. For more information, see the 
#       manual page for perlpod(1).
#

package Tag;

use strict;
use vars qw(%tags);

%tags=();

sub init {
	my ($file)=@_;
	open(INPUT,$file);
	while(defined ($_=<INPUT>)) {
		chop;
		if(/(\w*)\s+(.*)/) {
			$tags{$1}=$2;
		}
	}
}

1;
